"""Define a base client for interacting with Phyn."""
import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from typing import Optional

import boto3
from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError
from pycognito.aws_srp import AWSSRP

from .partners import KOHLER_API
from .mqtt import MQTTClient
from .device import Device
from .errors import BrandError, RequestError
from .home import Home


_LOGGER = logging.getLogger(__name__)

BRANDS = {
    'phyn': 0,
    'kohler': 1,
}

DEFAULT_HEADER_CONTENT_TYPE: str = "application/json"
DEFAULT_HEADER_USER_AGENT: str = "phyn/18 CFNetwork/1331.0.7 Darwin/21.4.0"
DEFAULT_HEADER_CONNECTION: str = "keep-alive"
DEFAULT_HEADER_API_KEY: list[str] = [
    "E7nfOgW6VI64fYpifiZSr6Me5w1Upe155zbu4lq8",
    "OOLFiYu7Ts5RKI4BV6WeI3zb38HU76vZ8x5lFX6Y",
]
DEFAULT_HEADER_ACCEPT: str = "application/json"
DEFAULT_HEADER_ACCEPT_ENCODING: str = "gzip, deflate, br"

COGNITO_REGION: str = "us-east-1"
COGNITO_POOL_ID: str = "us-east-1_UAv6IUsyh"
COGNITO_CLIENT_ID: str = "5q2m8ti0urmepg4lup8q0ptldq"

DEFAULT_TIMEOUT: int = 10


class API:
    """Define the API object."""

    def __init__(
        self, username: str, password: str, *, phyn_brand: str, session: Optional[ClientSession] = None,
        client_id: Optional[str] = None, verify_ssl: bool = True, proxy: Optional[str] = None,
        proxy_port: Optional[int] = None
    ) -> None:
        """Initialize."""
        if phyn_brand not in BRANDS:
            raise BrandError("Invalid phyn brand")

        self._brand: str = BRANDS[phyn_brand]

        self._username: str = username
        if self._brand != BRANDS['phyn']:
            self._password: str = None
            self._partner_api = None
            self._partner_password: str = password
            self._cognito: dict[str] = None
        else:
            self._password: str = password
            self._cognito: dict[str] = {
                "app_client_id": COGNITO_CLIENT_ID,
                "pool_id": COGNITO_POOL_ID,
                "region": COGNITO_REGION,
            }

        self._session: ClientSession = session
        self._iot_id = None
        self._iot_credentials = None
        self.mqtt = None
        self._id_token = None
        self._refresh_token = None
        self._mqtt_settings = {}

        self.verify_ssl = verify_ssl
        self.proxy = proxy
        self.proxy_port = proxy_port
        self.proxy_url: Optional[str] = None
        if self.proxy is not None and self.proxy_port is not None:
            self.proxy_url = f"https://{proxy}:{proxy_port}"

        self._token: Optional[str] = None
        self._token_expiration: Optional[datetime] = None

        self.home: Home = Home(self._request)
        self.device: Device = Device(self._request)
        self.mqtt = MQTTClient(self, client_id=client_id, verify_ssl=verify_ssl, proxy=proxy, proxy_port=proxy_port)

    @property
    def username(self) -> Optional[str]:
        """Get the API username"""
        return self._username

    async def _request(self, method: str, url: str, token_type:str = "access", **kwargs) -> dict:
        """Make a request against the API.

        :param method: GET or POST request
        :type method: str
        :param url: API URL
        :type url: str
        :param token_type: ID or Access token, defaults to "access"
        :type token_type: str, optional
        :raises RequestError: Error if issue accessing URL
        :return: JSON response
        :rtype: dict
        """
        if self._token_expiration and datetime.now() >= self._token_expiration:
            _LOGGER.info("Requesting new access token to replace expired one")

            # Nullify the token so that the authentication request doesn't use it:
            self._token = None

            # Nullify the expiration so the authentication request doesn't get caught
            # here:
            self._token_expiration = None

            await self.async_authenticate()

        kwargs.setdefault("headers", {})
        kwargs["headers"].update(
            {
                "Content-Type": DEFAULT_HEADER_CONTENT_TYPE,
                "User-Agent": DEFAULT_HEADER_USER_AGENT,
                "Connection": DEFAULT_HEADER_CONNECTION,
                "x-api-key": DEFAULT_HEADER_API_KEY[self._brand],
                "Accept": DEFAULT_HEADER_ACCEPT,
                "Accept-Encoding": DEFAULT_HEADER_ACCEPT_ENCODING,
            }
        )

        if token_type == "access":
            if self._token:
                kwargs["headers"]["Authorization"] = self._token
        elif token_type == "id":
            if self._id_token:
                kwargs["headers"]["Authorization"] = self._id_token

        if self.proxy_url is not None:
            kwargs["proxy"] = self.proxy_url

        if not self.verify_ssl:
            kwargs["ssl"] = False

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        try:
            async with session.request(method, url, **kwargs) as resp:
                data: dict = await resp.json(content_type=None)
                resp.raise_for_status()
                return data
        except ClientError as err:
            raise RequestError(f"There was an error while requesting {url}") from err
        finally:
            if not use_running_session:
                await session.close()

    async def async_authenticate(self) -> None:
        """Authenticate the user and set the access token with its expiration."""
        if self._brand == BRANDS["kohler"]:
            if self._password is None:
                _LOGGER.info("Auhenticating to Kohler")
                self._partner_api = KOHLER_API(self._username, self._partner_password, verify_ssl=self.verify_ssl,
                                               proxy=self.proxy, proxy_port=self.proxy_port)
                await self._partner_api.authenticate()
                self._password = self._partner_api.get_phyn_password()
                self._cognito = self._partner_api.get_cognito_info()
                self._mqtt_settings = self._partner_api.get_mqtt_info()

        executor = ThreadPoolExecutor()
        future = executor.submit(self._authenticate)
        auth_response = await asyncio.wrap_future(future)

        access_token = auth_response["AuthenticationResult"]["AccessToken"]
        expires_in = auth_response["AuthenticationResult"]["ExpiresIn"]
        id_token = auth_response["AuthenticationResult"]["IdToken"]
        refresh_token = auth_response["AuthenticationResult"]["RefreshToken"]

        self._token = access_token
        self._token_expiration = datetime.now() + timedelta(seconds=expires_in)
        self._id_token = id_token
        self._refresh_token = refresh_token

    def _authenticate(self):
        """boto3 is synchronous, so authenticate in a separate thread."""
        _LOGGER.info("Requesting token from AWS")
        client = boto3.client("cognito-idp", region_name=self._cognito['region'])
        aws = AWSSRP(
            username=self._username,
            password=self._password,
            pool_id=self._cognito['pool_id'],
            client_id=self._cognito['app_client_id'],
            client=client,
        )
        auth_response = aws.authenticate_user()
        return auth_response


async def async_get_api(
    username: str, password: str, *, phyn_brand: str = "phyn", session: Optional[ClientSession] = None,
    client_id: Optional[str] = None, verify_ssl: bool = True, proxy: Optional[str] = None,
    proxy_port: Optional[int] = None
) -> API:
    """Instantiate an authenticated API object.

    :param session: An ``aiohttp`` ``ClientSession``
    :type session: ``aiohttp.client.ClientSession``
    :param email: A Phyn email address
    :type email: ``str``
    :param password: A Phyn password
    :type password: ``str``
    :param phyn_brand: A brand for phyn
    :type phyn_brand: ``str``
    :param client_id: A MQTT client id name
    :type client_id: ``str``
    :param verify_ssl: Should SSL certificates be verified
    :type verify_ssl: ``bool``
    :param proxy: HTTP proxy hostname/IP
    :type proxy: ``str``
    :param proxy_port: Port for HTTP proxy
    :type proxy_port: ``int``
    :rtype: :meth:`aiophyn.api.API`
    """
    api = API(username, password, phyn_brand=phyn_brand, session=session, client_id=client_id,
              verify_ssl=verify_ssl, proxy=proxy, proxy_port=proxy_port)
    await api.async_authenticate()
    return api



""" Define Kohler Partner class """

import logging
import re
import uuid
import json
import base64
import binascii

from datetime import datetime, timedelta
from typing import Optional
from aiohttp import ClientSession, ClientTimeout, CookieJar


from Crypto.Cipher import AES
from Crypto.Util.Padding import unpad

_LOGGER = logging.getLogger(__name__)

DEFAULT_TIMEOUT: int = 10

class KOHLER_API:
    """API for Kohler to access Phyn Devices"""
    def __init__(
        self, username: str, password: str, verify_ssl: bool = True, proxy: Optional[str] = None,
        proxy_port: Optional[int] = None
    ):
        self._username: str = username
        self._password: str = password
        self._phyn_password: str = None
        self._user_id: str = None

        self._token: str = None
        self._token_expiration = None
        self._refresh_token = None
        self._refresh_token_expiration = None
        self._mobile_data = None

        self.verify_ssl = verify_ssl
        self.ssl = False if verify_ssl is False else None
        self.proxy = proxy
        self.proxy_port = proxy_port
        self.proxy_url: Optional[str] = None
        if self.proxy is not None and self.proxy_port is not None:
            self.proxy_url = f"https://{proxy}:{proxy_port}"

        self._session: ClientSession = None

    def get_cognito_info(self):
        """Get cognito information"""
        return self._mobile_data['cognito']

    def get_mqtt_info(self):
        """Get MQTT url"""
        return self._mobile_data['wss']

    def get_phyn_password(self):
        """Get phyn password"""
        return self._phyn_password

    async def authenticate(self):
        """Authenticate with Kohler and Phyn"""
        use_running_session = self._session and not self._session.closed
        if not use_running_session:
            self._session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT), cookie_jar=CookieJar(quote_cookie=False))

        await self.b2c_login()
        token = await self.get_phyn_token()
        await self._session.close()
        self._phyn_password = await self.token_to_password(token)

    async def b2c_login(self):
        """Login to Kohler"""
        _LOGGER.debug("Logging into Kohler")
        client_request_id = str(uuid.uuid4())

        # Get CSRF token and initialize values
        params = {
          "response_type": "code",
          "client_id": "8caf9530-1d13-48e6-867c-0f082878debc",
          "client-request-id": client_request_id,
          "scope": "https%3A%2F%2Fkonnectkohler.onmicrosoft.com%2Ff5d87f3d-bdeb-4933-ab70-ef56cc343744%2Fapiaccess%20" + 
            "openid%20offline_access%20profile",
          "redirect_uri": "msauth%3A%2F%2Fcom.kohler.hermoth%2F2DuDM2vGmcL4bKPn2xKzKpsy68k%253D",
          "prompt": "login",
        }
        get_vars = '&'.join([ f"{x[0]}={x[1]}" for x in params.items() ])
        resp = await self._session.get(
            'https://konnectkohler.b2clogin.com/tfp/konnectkohler.onmicrosoft.com/B2C_1A_signin/oAuth2/v2.0/authorize?'
            + get_vars, ssl=self.ssl, proxy=self.proxy_url)
        match = re.search(r'"(StateProperties=([a-zA-Z0-9]+))"', await resp.text())
        state_properties = match.group(1)

        cookies = self._session.cookie_jar.filter_cookies('https://konnectkohler.b2clogin.com')
        csrf = None
        for key, cookie in cookies.items():
            if key == "x-ms-cpim-csrf":
                csrf = cookie.value

        # Login
        headers = {
            "X-CSRF-TOKEN": csrf,
        }
        login_vars = {
            "request_type": "RESPONSE",
            "signInName": self._username,
            "password": self._password,
        }
        resp = await self._session.post("https://konnectkohler.b2clogin.com/konnectkohler.onmicrosoft.com/" +
                                        "B2C_1A_signin/SelfAsserted?p=B2C_1A_signin&" + state_properties,
                                        headers=headers, data=login_vars, ssl=self.ssl, proxy=self.proxy_url)

        params = {
            "rememberMe": "false",
            "csrf_token": csrf,
            "tx": state_properties,
            "p": "B2C_1A_signin"
        }
        args = '&'.join([ f"{x[0]}={x[1]}" for x in params.items() ])
        resp = await self._session.get("https://konnectkohler.b2clogin.com/konnectkohler.onmicrosoft.com/" +
                                       "B2C_1A_signin/api/CombinedSigninAndSignup/confirmed?" + args,
                                       allow_redirects=False, ssl=self.ssl, proxy=self.proxy_url)
        
        if 'Location' not in resp.headers:
            raise Exception("Unable to login to Kohler")
        matches = re.search(r'code=([^&]+)', resp.headers['Location'])
        code = matches.group(1)

        # Get tokens
        headers = {
            "x-app-name": "com.kohler.hermoth",
            "x-app-ver": "2.7",
        }
        params = {
            "client-request-id": client_request_id,
            "client_id": "8caf9530-1d13-48e6-867c-0f082878debc",
            "client_info": "1",
            "x-app-name": "com.kohler.hermoth",
            "x-app-ver": "2.7",
            "redirect_uri": "msauth://com.kohler.hermoth/2DuDM2vGmcL4bKPn2xKzKpsy68k%3D",
            "scope": "https://konnectkohler.onmicrosoft.com/f5d87f3d-bdeb-4933-ab70-ef56cc343744/apiaccess" + 
                " openid offline_access profile",
            "grant_type": "authorization_code",
            "code": code,
        }
        resp = await self._session.post("https://konnectkohler.b2clogin.com/tfp/konnectkohler.onmicrosoft.com/" +
                                        "B2C_1A_signin/%2FoAuth2%2Fv2.0%2Ftoken", data=params, ssl=self.ssl,
                                        proxy=self.proxy_url)

        data = await resp.json()
        if "client_info" not in data:
            await self._session.close()
            raise Exception("Unable to get client data")

        client_info = json.loads(base64.b64decode(data['client_info'] + '==').decode())
        self._user_id = re.sub('-b2c_1a_signin$', '', client_info['uid'])
        #await self.home.set_user_id(self._uid)

        self._token = data['access_token']
        self._token_expiration = datetime.now() + timedelta(seconds=data['expires_in'])
        self._refresh_token = data['refresh_token']
        self._refresh_token_expiration = datetime.now() + timedelta(seconds=data['refresh_token_expires_in'])
        _LOGGER.debug("Received Kohler Token")

    async def get_phyn_token(self):
        """ Get a phyn access token"""
        params = {
          "partner": "kohler",
          "partner_user_id": self._user_id,
          "email": self._username,
        }
        args = "&".join([ f"{x[0]}={x[1]}" for x in params.items() ])
        headers = {
          "Accept": "application/json",
          "Accept-encoding": "gzip",
          "Authorization": f"Bearer partner-{self._token}",
          "Content-Type": "application/json",
          "User-Agent": "okhttp/4.10.0"
        }

        _LOGGER.info("Getting Kohler settings from Phyn")
        resp = await self._session.get(f"https://api.phyn.com/settings/app/com.kohler.mobile?{args}", headers=headers,
                                       ssl=self.ssl, proxy=self.proxy_url)
        mobile_data = await resp.json()
        if "error_msg" in mobile_data:
            await self._session.close()
            raise Exception(f"Kohler {mobile_data['error_msg']}")

        if "cognito" not in mobile_data:
            await self._session.close()
            raise Exception("Unable to find cognito information")
        self._mobile_data = mobile_data

        _LOGGER.debug("Getting token from Phyn")
        params = {
          "email": self._username,
          "partner": "kohler",
          "partner_user_id": self._user_id
        }
        args = "&".join([ f"{x[0]}={x[1]}" for x in params.items() ])
        headers = {
          "Accept": "application/json, text/plain, */*",
          "Accept-encoding": "gzip",
          "Authorization": f"Bearer partner-{self._token}",
          "Content-Type": "application/json",
          "x-api-key": mobile_data['pws_api']['app_api_key']
        }
        resp = await self._session.get(f"https://api.phyn.com/partner-user-setup/token?{args}", headers=headers,
                                       ssl=self.ssl, proxy=self.proxy_url)
        data = await resp.json()
        if "token" not in data:
            await self._session.close()
            raise Exception("Token not found")
        _LOGGER.debug("Token received")
        return data['token']

    async def token_to_password(self, token):
        """Convert a token to a Phyn password"""
        b64hex = base64.b64decode((token + '=' * (5 - (len(token) % 4))).replace('_','/').replace('-','+')).hex()

        try:
            keydata = binascii.hexlify(base64.b64decode(self._mobile_data['partner']['comm_id'])).decode()
        except Exception as e:
            raise Exception("Error getting password decryption key") from e

        key = keydata[32:]
        iv = b64hex[18:(18+32)]
        ct = b64hex[50:(len(b64hex)-64)]
        cipher = AES.new(bytes.fromhex(key), AES.MODE_CBC, iv=bytes.fromhex(iv))
        return unpad(cipher.decrypt(bytearray.fromhex(ct)), AES.block_size).decode()



import asyncio
import logging
from pprint import pprint

from aiohttp import ClientSession

from aiophyn import async_get_api
from aiophyn.errors import PhynError

_LOGGER = logging.getLogger()


async def device_dump(username: str, password: str) -> None:
    device_states = []
    away_mode_states = []
    async with ClientSession() as session:
        try:
            api = await async_get_api(username, password, session=session)
            all_homes = await api.home.get_homes(username)
            for home in all_homes:
                for device_id in home.get("device_ids", []):
                    device_state = await api.device.get_state(device_id)
                    device_states.append(device_state)

                    away_mode = await api.device.get_away_mode(device_id)
                    away_mode_states.append(away_mode)
            print("\n" * 3)
            pprint(device_states)
            print("\n" * 3)
            pprint(away_mode_states)
            print("\n" * 3)

        except PhynError as err:
            _LOGGER.error("There was an error: %s", err)


async def main() -> None:
    username = input("Username: ")
    password = input("Password: ")
    await device_dump(username, password)


asyncio.run(main())



""" Module providing a MQTT provider """

import asyncio
import logging

from typing import Any, Dict, Union, Optional

import inspect
import json
import time
import urllib
import ssl
import socket
import re
import socks
import paho.mqtt.client as paho_mqtt

from .const import API_BASE

_LOGGER = logging.getLogger(__name__)

class AIOHelper:
    """Helper class for Asynchronous IO"""
    def __init__(self, client: paho_mqtt.Client) -> None:
        self.loop = asyncio.get_running_loop()
        self.client = client
        self.client.on_socket_open = self._on_socket_open
        self.client.on_socket_close = self._on_socket_close
        self.client._on_socket_register_write = self._on_socket_register_write
        self.client._on_socket_unregister_write = \
            self._on_socket_unregister_write
        self.misc_task: Optional[asyncio.Task] = None

    def _on_socket_open(self,
                        client: paho_mqtt.Client,
                        userdata: Any,
                        sock: socket.socket
                        ) -> None:
        # pylint: disable=unused-argument
        _LOGGER.info("MQTT Socket Opened")
        self.loop.add_reader(sock, client.loop_read)
        self.misc_task = self.loop.create_task(self.misc_loop())

    def _on_socket_close(self, client: paho_mqtt.Client, userdata: Any, sock: socket.socket) -> None:
        # pylint: disable=unused-argument
        _LOGGER.info("MQTT Socket Closed")
        self.loop.remove_reader(sock)
        if self.misc_task is not None:
            self.misc_task.cancel()

    def _on_socket_register_write(self,
                                  client: paho_mqtt.Client,
                                  userdata: Any,
                                  sock: socket.socket
                                  ) -> None:
        # pylint: disable=unused-argument
        self.loop.add_writer(sock, client.loop_write)

    def _on_socket_unregister_write(self,
                                    client: paho_mqtt.Client,
                                    userdata: Any,
                                    sock: socket.socket
                                    ) -> None:
        # pylint: disable=unused-argument
        self.loop.remove_writer(sock)

    async def misc_loop(self) -> None:
        """Loop for MQTT"""
        while self.client.loop_misc() == paho_mqtt.MQTT_ERR_SUCCESS:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break
        _LOGGER.info("MQTT Misc Loop Complete")

class Timer:
    """ Class to run a job with a timeout """
    def __init__(self, callback):
        _LOGGER.info("Creating timer")
        self._timeout = 0
        self._callback = callback
        self._task = None

    async def _job(self, timeout):
        """ Run the job with a timeout """
        await asyncio.sleep(timeout)
        _LOGGER.debug("Executing timer callback")
        if inspect.iscoroutinefunction(self._callback):
            await self._callback()
        else:
            self._callback()

    def cancel(self):
        """ Cancel a timer task """
        if self._task is not None:
            self._task.cancel()
            self._task = None

    def start(self, timeout):
        """ Start a timer task """
        if self._task is not None:
            self._task.cancel()
        _LOGGER.debug("Starting timer job for %s seconds", timeout)
        self._task = asyncio.create_task(self._job(timeout))

class MQTTClient:
    """AIO MQTT client """
    def __init__(self, api, client_id: str =None, verify_ssl: bool =True, proxy: str =None, proxy_port: int =None):
        self.event_loop = asyncio.get_running_loop()
        self.api = api
        self.pending_acks = {}
        self.topics = []
        self.connect_evt: asyncio.Event = asyncio.Event()
        self.connect_task = None
        self.disconnect_evt: Optional[asyncio.Event] = None
        self.reconnect_evt: asyncio.Event = asyncio.Event()
        self.host = None 
        self.port = 443

        if client_id is None:
            client_id = "aiophyn-%s" % int(time.time())

        self.client = paho_mqtt.Client(client_id=client_id, transport="websockets")
        self.helper: AIOHelper = None
        self.reconnect_timer = Timer(self._process_reconnect)

        self.verify_ssl: bool = verify_ssl
        self.proxy: Optional[str] = proxy
        self.proxy_port: Optional[int] = proxy_port

        self.client.on_connect = self._on_connect
        self.client.on_disconnect = self._on_disconnect
        self.client.on_subscribe = self._on_subscribe
        self.client.on_message = self._on_message

        self._handlers = {
            "connect": [],
            "disconnect": [],
            "update": []
        }

    async def add_event_handler(self, type, target):
        """Add an event handler for MQTT events"""
        if type not in self._handlers.keys():
            return False

        if target in self._handlers[type]:
            return True
        self._handlers[type].append(target)

    async def connect(self):
        """ Create a conenction to the MQTT server """
        self.host, path = await self.get_mqtt_info()
        self.client.ws_set_options(path, headers={'Host': self.host})

        if self.verify_ssl:
            context = ssl.SSLContext()
            self.client.tls_set_context(context)
        else:
            context = ssl.SSLContext()
            context.verify_mode = ssl.CERT_NONE
            context.check_hostname = False
            self.client.tls_set_context(context)
            self.client.tls_insecure_set(True)

        if self.proxy is not None and self.proxy_port is not None:
            self.client.proxy_set(proxy_type=socks.HTTP, proxy_addr=self.proxy, proxy_port=self.proxy_port)

        self.helper = AIOHelper(self.client)
        _LOGGER.info("Connecting to mqtt websocket: %s", self.host)
        self.reconnect_timer.start(5)
        await self.event_loop.run_in_executor(
                None,
                self.client.connect,
                self.host,
                self.port,
            )
    
    def disconnect(self):
        """Disconnect from server"""
        self.disconnect_evt = asyncio.Event()
        _LOGGER.info("MQTT client disconnecting...")
        self.client.disconnect()
    
    async def disconnect_and_wait(self):
        """Disconnect from server and wait"""
        self.disconnect()
        await self.disconnect_evt.wait()

    async def get_mqtt_info(self):
        """ Gets WebSocket URL and parameters for a MQTT connection
            Returns a list of url and path
        """
        user_id = urllib.parse.quote_plus(self.api.username)
        try:
            wss_data = await self.api._request("post", f"{API_BASE}/users/{user_id}/iot_policy", token_type="id")
        except:
            Exception("Could not get WebSocket/MQTT url from API")

        match = re.match(r'wss:\/\/([a-zA-Z0-9\.\-]+)(\/mqtt?.*)', wss_data['wss_url'])
        if not match:
            raise Exception("Could not find WebSocket/MQTT url")

        return match.group(1), match.group(2)


    async def subscribe(self, topic):
        """Subscribe to a MQTT topic"""
        _LOGGER.info("Attempting to subscribe to: %s", topic)
        res, msg_id = self.client.subscribe(topic, 0)
        self.pending_acks[msg_id] = topic


    def _on_connect(self,
                    client: paho_mqtt.Client,
                    user_data: Any,
                    flags: Dict[str, Any],
                    reason_code: Union[int, paho_mqtt.ReasonCodes],
                    properties: Optional[paho_mqtt.Properties] = None
                    ) -> None:
        # pylint: disable=unused-argument
        _LOGGER.info("MQTT Client Connected")
        if reason_code == 0:
            _LOGGER.info("Trying to run timer...")
            self.reconnect_timer.cancel()
            self.reconnect_timer.start(3600)
            self.connect_evt.set()
        else:
            if isinstance(reason_code, int):
                err_str = paho_mqtt.connack_string(reason_code)
            else:
                err_str = reason_code.getName()
            _LOGGER.info("MQTT Connection Failed: %s", err_str)

    def _on_disconnect(self,
                       client: paho_mqtt.Client,
                       user_data: Any,
                       reason_code: int,
                       properties: Optional[paho_mqtt.Properties] = None
                       ) -> None:
        # pylint: disable=unused-argument
        if self.disconnect_evt is not None:
            self.disconnect_evt.set()
            _LOGGER.info("Client disconnected, not attempting to reconnect")
        elif not self.is_connected():
            # The server connection was dropped, attempt to reconnect
            _LOGGER.info("MQTT Server Disconnected, reason: %s", paho_mqtt.error_string(reason_code))
            self.reconnect_timer.cancel()
            if self.connect_task is None:
                self.connect_task = asyncio.create_task(self._do_reconnect(True))
        self.connect_evt.clear()

    def is_connected(self) -> bool:
        """ Checks if the client is connected """
        return self.client.is_connected()

    async def _process_reconnect(self):
        #self.connect_task = True
        _LOGGER.info("Processing reconnect request")

        self.disconnect_evt = asyncio.Event()
        if self.is_connected():
            self.client.disconnect()
            await self.disconnect_evt.wait()

        self.connect_task = asyncio.create_task(self._do_reconnect(True))

    async def _do_reconnect(self, first: bool = False) -> None:
        if self.reconnect_evt.is_set():
            _LOGGER.info("Already attempting to reconnect, second attemp cancelled.")
            return

        _LOGGER.info("Attempting MQTT Connect/Reconnect")
        self.reconnect_evt.set()
        last_err: Exception = Exception()
        connect_attempts = 0
        t: float = 2.
        while True:
            if not first:
                try:
                    if connect_attempts > 6:
                        t = 60.
                    elif connect_attempts > 3:
                        t = 10.
                    _LOGGER.debug("MQTT throttle for %s seconds", t)
                    await asyncio.sleep(t)
                except asyncio.CancelledError:
                    self.reconnect_evt.clear()
                    raise
            first = False
            connect_attempts += 1
            try:
                self.host, path = await self.get_mqtt_info()
                self.client.ws_set_options(path, headers={'Host': self.host})
                _LOGGER.info("Attempting to reconnnect...")
                await self.event_loop.run_in_executor(
                        None,
                        self.client.connect,
                        self.host,
                        self.port,
                    )

                await asyncio.wait_for(self.connect_evt.wait(), timeout=2.)
                if not self.connect_evt.is_set():
                    _LOGGER.info("Timeout while waiting for MQTT connection")
                    continue

                # Re-subscribe to all topics
                topics = list(set(self.topics))
                tasks = [self.subscribe(topic) for topic in topics]
                await asyncio.gather(*tasks)
            except asyncio.CancelledError:
                self.reconnect_evt.clear()
                raise
            except Exception as e:
                if type(last_err) is not type(e) or last_err.args != e.args:
                    _LOGGER.warning("MQTT Connection Error")
                    last_err = e
                continue
            break
        self.reconnect_evt.clear()
        self.disconnect_evt = None
        self.connect_task = None

    def _on_message(
        self, client: paho_mqtt.Client, userdata: Any, message: paho_mqtt.MQTTMessage
    ) -> None:
        # pylint: disable=unused-argument
        msg = message.payload.decode()
        _LOGGER.debug("Message received on %s: %s", message.topic, msg)
        try:
            data = json.loads(msg)
        except json.decoder.JSONDecodeError:
            _LOGGER.info("Received invalid JSON message: %s", msg)

        if message.topic.startswith("prd/app_subscriptions/"):
            device_id = message.topic.split('/')[2]
        else:
            device_id = None

        for h in self._handlers["update"]:
            asyncio.ensure_future(h(device_id, data))

    def _on_subscribe(
        self,
        client: paho_mqtt.Client,
        userdata: Any,
        mid: int,
        granted_qos: tuple[int] | list[paho_mqtt.ReasonCodes],
        properties: paho_mqtt.Properties | None = None,
    ) -> None:
        # pylint: disable=unused-argument
        if mid in self.pending_acks:
            _LOGGER.info("Subscribed to: %s", self.pending_acks[mid])
            self.topics.append(self.pending_acks[mid])
            del self.pending_acks[mid]
        else:
            _LOGGER.info("Subscribed: %s %s %s", userdata, str(mid), str(granted_qos))



"""Define package errors."""


class PhynError(Exception):
    """Define a base error."""

    ...


class RequestError(PhynError):
    """Define an error related to invalid requests."""

    ...

class BrandError(Exception):
    """ Define an error related to invalid brands."""

    ...



"""Define /home endpoints."""
from typing import Awaitable, Callable

from .const import API_BASE


class Home:
    """Define an object to handle the endpoints."""

    def __init__(self, request: Callable[..., Awaitable]) -> None:
        """Initialize."""
        self._request: Callable[..., Awaitable] = request

    async def get_homes(
        self,
        user_id: str,
    ) -> dict:
        """Return info for all homes.

        :param user_id: Phyn username (email)
        :type user_id: ``str``
        :rtype: ``list``
        """
        params = {"user_id": user_id}

        return await self._request("get", f"{API_BASE}/homes", params=params)



"""Define /devices endpoints."""
from typing import Awaitable, Any, Callable, Optional

from .const import API_BASE


class Device:
    """Define an object to handle the endpoints."""

    def __init__(self, request: Callable[..., Awaitable]) -> None:
        """Initialize."""
        self._request: Callable[..., Awaitable] = request

    async def get_state(self, device_id: str) -> dict:
        """Return state of a device.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        return await self._request("get", f"{API_BASE}/devices/{device_id}/state")

    async def get_consumption(
        self,
        device_id: str,
        duration: str,
        precision: int = 6,
        details: Optional[str] = False,
        event_count: Optional[str] = False,
        comparison: Optional[str] = False,
    ) -> dict:
        """Return water consumption of a device.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :param duration: Date string formatted as 'YYYY/MM/DD', 'YYYY/MM', or 'YYYY'
        :type duration: ``str``
        :param precision: Decimal places of measurement precision
        :type precision: ``int``
        :param details: Include detailed breakdown of consumption
        :type details: ``bool``
        :param event_count: Include the event count
        :type event_count: ``bool``
        :param comparison: Include comparison data
        :type comparison: ``bool``
        :rtype: ``dict``
        """

        params = {
            "device_id": device_id,
            "duration": duration,
            "precision": precision,
        }

        if details:
            params["details"] = "Y"

        if event_count:
            params["event_count"] = "Y"

        if comparison:
            params["comparison"] = "Y"

        return await self._request(
            "get", f"{API_BASE}/devices/{device_id}/consumption/details", params=params
        )

    async def get_water_statistics(self, device_id: str, from_ts, to_ts):
        """Get statistics about a PW1 sensor

        :param device_id: Unique identifier for the device
        :type device_id: str
        :param from_ts: Lower bound timestamp. This is a timestamp with thousands as integer
        :type from_ts: int
        :param to_ts: Upper bound timestamp. This is a timestamp with thousands as integer
        :type to_ts: int
        :return: List of dictionaries of results. 
        :rtype: List[dict[str, Any]]
        """
        params = {
            "from_ts": from_ts,
            "to_ts": to_ts
        }

        return await self._request(
            "get", f"{API_BASE}/devices/{device_id}/water_statistics/history/", params=params
        )

    async def open_valve(self, device_id: str) -> None:
        """Open a device shutoff valve.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        return await self._request(
            "post",
            f"{API_BASE}/devices/{device_id}/sov/Open",
        )

    async def close_valve(self, device_id: str) -> None:
        """Close a device shutoff valve.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        return await self._request(
            "post",
            f"{API_BASE}/devices/{device_id}/sov/Close",
        )

    async def get_away_mode(self, device_id: str) -> dict:
        """Return away mode status of a device.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        return await self._request("get", f"{API_BASE}/preferences/device/{device_id}/leak_sensitivity_away_mode")


    async def enable_away_mode(self, device_id: str) -> None:
        """Enable the device's away mode.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        data = [
            {
                "name": "leak_sensitivity_away_mode",
                "value": "true",
                "device_id": device_id,
            }
        ]
        return await self._request(
            "post", f"{API_BASE}/preferences/device/{device_id}", json=data
        )

    async def disable_away_mode(self, device_id: str) -> None:
        """Disable the device's away mode.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        data = [
            {
                "name": "leak_sensitivity_away_mode",
                "value": "false",
                "device_id": device_id,
            }
        ]
        return await self._request(
            "post", f"{API_BASE}/preferences/device/{device_id}", json=data
        )
    
    async def get_autoshuftoff_status(self, device_id: str) -> dict:
        """Get phyn device preferences.

        :param device_id: Unique identifier for the device
        :type device_id: str
        :return: List of dicts with the following keys: created_ts, device_id, name, updated_ts, value
        :rtype: dict
        """
        return await self._request(
            "get", f"{API_BASE}/devices/{device_id}/auto_shutoff"
        )
    

    async def get_device_preferences(self, device_id: str) -> dict:
        """Get phyn device preferences.

        :param device_id: Unique identifier for the device
        :type device_id: str
        :return: List of dicts with the following keys: created_ts, device_id, name, updated_ts, value
        :rtype: dict
        """
        return await self._request(
            "get", f"{API_BASE}/preferences/device/{device_id}"
        )
    
    async def get_health_tests(self, device_id: str) -> dict:
        """Get phyn device preferences.

        :param device_id: Unique identifier for the device
        :type device_id: str
        :return: List of dicts with the following keys
        :rtype: dict
        """
        return await self._request(
            "get", f"{API_BASE}/devices/{device_id}/health_tests?list_type=grouped"
        )
    
    async def get_latest_firmware_info(self, device_id: str) -> dict:
        """Get Latest Firmware Information

        :param device_id: Unique identifier for the device
        :type device_id: str
        :return: Returns dict with fw_img_name, fw_version, product_code
        :rtype: dict
        """
        return await self._request(
            "get", f"{API_BASE}/firmware/latestVersion/v2?device_id={device_id}"
        )

    async def run_leak_test(self, device_id: str, extended_test: bool = False):
        """Run a leak test

        :param device_id: Unique identifier for the device
        :type device_id: str
        :param extended_test: True if the test be extended, defaults to False
        :type extended_test: bool, optional
        """
        data = {
            "initiator": "App",
            "test_duration": "e" if extended_test is True else "s"
        }
        return await self._request(
            "post", f"{API_BASE}/devices/{device_id}/health_tests", json=data
        )

    async def set_autoshutoff_enabled(self, device_id: str, shutoff_on: bool, time: int | None = None) -> None:
        """Set autoshutoff enabled

        :param device_id: Unique identifier for the device
        :type device_id: str
        :param shutoff_on: Turn autoshutoff on (True) or off (False). If false, also turn off for amount of time
        :type shutoff_on: bool
        :param time: Time for shutoff in seconds if disabling (30, 3600, 21600, 86400), or blank for indefinite
        :type time: int | None
        :param data: List of dicts which have the keys: device_id, name, value
        :type data: List[dict]
        """
        url = f"{API_BASE}/devices/{device_id}/auto_shutoff/status/"
        if shutoff_on == True:
            url += "Enable"
        else:
            url += "Disable"
            if time != None:
                url += "/%s" % time
        return await self._request(
            "post", url
        )

    async def set_device_preferences(self, device_id: str, data: list[dict]) -> None:
        """Set device preferences

        :param device_id: Unique identifier for the device
        :type device_id: str
        :param data: List of dicts which have the keys: device_id, name, value
        :type data: List[dict]
        """
        return await self._request(
            "post", f"{API_BASE}/preferences/device/{device_id}", json=data
        )
